/*******************************************************************************
Gameinfoplus a library to get information from dedicated game servers.

Copyright (C) 2023  Jan Koester jan.koester@tuxist.de

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*******************************************************************************/

#include <iostream>
#include <algorithm>
#include <cstring>

#include <netplus/exception.h>

#include "hldsview.h"

#define A2S_HEADER_INFO 0x54
#define A2S_INFO_PAYLOAD_LEN_WITH_CHALLENGE    29
#define A2S_INFO_PAYLOAD_LEN_WITHOUT_CHALLENGE (A2S_INFO_PAYLOAD_LEN_WITH_CHALLENGE - 4)
#define A2S_INFO_PAYLOAD_CHALLENGE_OFFSET      (A2S_INFO_PAYLOAD_LEN_WITH_CHALLENGE - 4)

const uint8_t payload_template[A2S_INFO_PAYLOAD_LEN_WITH_CHALLENGE] = {
    0xFF, 0xFF, 0xFF, 0xFF, A2S_HEADER_INFO,
    'S', 'o', 'u', 'r', 'c', 'e', ' ', 'E', 'n', 'g', 'i', 'n', 'e', ' ', 'Q', 'u', 'e', 'r', 'y', '\0'
};

HldsView::HldsView(const char *addr,int port){
    try{
        serverSocket=std::make_shared<netplus::udp>(addr,port,1,0);
        clientSocket=std::make_shared<netplus::udp>();

        sport = port;
        serverSocket->connect(clientSocket);
        memset(challange,0,sizeof(challange));
    }catch(netplus::NetException &exp){
        throw exp.what();
    }
}

HldsView::~HldsView(){
}

HldsView::HldsView(const HldsView &view){
        serverSocket=std::make_shared<netplus::udp>(*view.serverSocket);
        clientSocket=std::make_shared<netplus::udp>();
        memcpy(challange,view.challange,sizeof(challange));
        serverSocket->connect(clientSocket);
        sport = view.sport;
};

void HldsView::refresh(HldsData &info){
    const std::lock_guard<std::mutex> lock(rlock);
    char buffer[1400];
    size_t rcv=0;
    try{
        uint8_t payload[A2S_INFO_PAYLOAD_LEN_WITH_CHALLENGE];
        memcpy(payload, payload_template, A2S_INFO_PAYLOAD_LEN_WITHOUT_CHALLENGE);

        serverSocket->sendData(clientSocket,(void*)&payload,A2S_INFO_PAYLOAD_LEN_WITHOUT_CHALLENGE);

        char recv[9];
        rcv=serverSocket->recvData(clientSocket,&recv,sizeof(recv));

        if(recv[4] == 0x41){
            memcpy(challange,recv+5,4);
        }

        AGAIN:
        if(challange[0]!=0){
            memcpy(payload + A2S_INFO_PAYLOAD_CHALLENGE_OFFSET, &challange, sizeof (challange));
            serverSocket->sendData(clientSocket,(void*)payload,A2S_INFO_PAYLOAD_LEN_WITH_CHALLENGE);
        }else{
            serverSocket->sendData(clientSocket,(void*)payload,A2S_INFO_PAYLOAD_LEN_WITHOUT_CHALLENGE);
        }


        rcv=serverSocket->recvData(clientSocket,buffer,1400);

        if(rcv >4 && buffer[4] == 0x41){
            memcpy(challange,buffer+5,4);
            goto AGAIN;
        }
        _parse(info,buffer,rcv);
    }catch(netplus::NetException &exp){
        throw exp.what();
    }
}

void HldsView::_parse(HldsData &info,const char* data, size_t len){
    info.Port=sport;
    info.header=0x00;

    for(int i=0; i<len; ++i){
        if(data[i]==0x49){
            info.header=0x49;
            data+=++i;
            len-=i;
        }
    }


    if(info.header!=0x49){
        throw "_parse: no hlds server!";
    }

    info.protocol=data[0];
    data=data+1;
    len-=1;

    for(int i=0; i<len; ++i){
        if(data[i]=='\0'){
            info.ServerName.assign(data,data+i);
            data+=++i;
            len-=i;
            break;
        }
    }

    for(int i=0; i<len; ++i){
        if(data[i]=='\0'){
            info.MapName.assign(data,data+i);
            data+=++i;
            len-=i;
            break;
        }
    }

    for(int i=0; i<len; ++i){
        if(data[i]=='\0'){
            info.ModName.assign(data,data+i);
            data+=++i;
            len-=i;
            break;
        }
    }


    for(int i=0; i<len; ++i){
        if(data[i]=='\0'){
            info.GameName.assign(data,data+i);
            data+=++i;
            len-=i;
            break;
        }
    }

    memcpy(&info.steamid,data,sizeof(short));
    data=data+sizeof(short);
    len-=sizeof(short);

    info.Players=data[0];
    data=data+1;
    len-=1;

    info.MaxPlayers=data[0];
    data=data+1;
    len-=1;

    info.BotsAmount=data[0];
    data=data+1;
    len-=1;

    info.servertype=data[0];
    data=data+1;
    len-=1;

    info.env=data[0];
    data=data+1;
    len-=1;

    if(data[0]==1)
        info.pwProtected=true;
    else
        info.pwProtected=false;
    data=data+1;
    len-=1;

    info.vac=data[0];
    data=data+1;
    len-=1;
}
